/*
 * debug - main.c
 * 	version 0.0.0
 * 	
 * Copyright (C) 2005  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: main.c,v 1.6.2.1 2005/11/29 17:38:32 krzysiek Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <errno.h>
#include <debug_io.h>
#include <debug_mem.h>
#include <getopt.h>

#define PROMPT	"> "

static int ac_compat = 0;	/* Linux MiniDEBUG compatibility mode */
static int parambase = 16;	/* used in Linux MiniDEBUG compatibility mode */
static int dump_align = 0;
static int rel_mem = 0;
static int rel_io = 0;
static unsigned long membase = 0;
static unsigned long iobase = 0;
static int hide_nonascii = 1;
static int hide_nonprintable = 1;

char *rl_gets(void)
{
	char *tmp;

	tmp = readline(PROMPT);
	if (tmp && *tmp)
		add_history(tmp);
	return tmp;	
}

void print_int8(uint8_t value)
{
	printf("0x%02hhx (%hhi)\n", value, value);
}
	
void print_int16(uint16_t value)
{
	printf("0x%04hx (%hi)\n", value, value);
}

void print_int32(uint32_t value)
{
	printf("0x%08x (%i)\n", value, value);
}

unsigned long gen_stoul(const char *str)
{
	if (ac_compat) {
		if (str[0] == '$')
			return (unsigned long) strtoll(str + 1, NULL, 16);
		if (str[0] == '#')
			return (unsigned long) strtoll(str + 1, NULL, 10);
		return (unsigned long) strtoll(str, NULL, parambase);
	}
	return (unsigned long) strtoll(str, NULL, 0);
}

int gen_parse_1ul(const char *args, unsigned long *addr)
{
	char cmd[64], addr_s[16];
	
	sscanf(args, "%64s %16s", cmd, addr_s);
	addr_s[15] = '\0';
	*addr = gen_stoul(addr_s);
	return 0;
}

int gen_parse_2ul(const char *args, unsigned long *addr, unsigned long *value)
{
	char cmd[64], addr_s[16], value_s[16];
	
	sscanf(args, "%64s %16s %16s", cmd, addr_s, value_s);
	addr_s[15] = '\0';
	value_s[15] = '\0';
	*addr = gen_stoul(addr_s);
	*value = gen_stoul(value_s);
	return 0;
}

int gen_parse_3ul(const char *args, unsigned long *addr, unsigned long *value,
		unsigned long *value2)
{
	char cmd[64], addr_s[24], value_s[24], value2_s[24];
	
	sscanf(args, "%64s %24s %24s %24s", cmd, addr_s, value_s, value2_s);
	addr_s[23] = '\0';
	value_s[23] = '\0';
	value2_s[23] = '\0';
	*addr = gen_stoul(addr_s);
	*value = gen_stoul(value_s);
	*value2 = gen_stoul(value2_s);
	return 0;
}

static char c_to_pc(uint8_t c)
{
	if ((hide_nonprintable && (c < 32)) || (hide_nonascii && (c > 127)))
		return '.';
	return c;
}

void dump_area_a(unsigned long base, uint8_t *buff, size_t len)	
{
	unsigned long i, j;

	for (i = base & ~0xf; i < base + len; i += 16 ) {
		printf("0x%08lx: ", i);
		for (j = 0; j < 16; j++) {
			if (j == 8)
				printf(" ");

			if ((i + j < base) || (i + j >= base + len))
				printf("   ");
			else
				printf("%02hhx ", buff[i + j - base]);
		}
		printf(" ");
		for (j = 0; j < 16; j++) {
			if ((i + j < base) || (i + j >= base + len))
				printf(" ");
			else
				printf("%c", c_to_pc(buff[i + j - base]));
		}
		printf("\n");
	}
}

void dump_area(unsigned long base, uint8_t *buff, size_t len)	
{
	unsigned long i, j;

	for (i = base; i < base + len; i += 16 ) {
		printf("0x%08lx: ", i);
		for (j = 0; j < 16; j++) {
			if (j == 8)
				printf(" ");

			if (i + j >= base + len)
				printf("   ");
			else
				printf("%02hhx ", buff[i + j - base]);
		}
		printf(" ");
		for (j = 0; j < 16; j++) {
			if ((i + j < base) || (i + j >= base + len))
				printf(" ");
			else
				printf("%c", c_to_pc(buff[i + j - base]));
		}
		printf("\n");
	}
}

int cmd_inb(const char *args, int rel)
{
	unsigned long port;
	uint8_t value;
	int ret;

	gen_parse_1ul(args, &port);
	if (rel)
		port += iobase;
	ret = d_inb(port, &value);
	if (ret) {
		fprintf(stderr, "d_inb error: %s\n", strerror(-ret));
		return ret;
	}
	print_int8(value);
	return 0;
}

int cmd_outb(const char *args, int rel)
{
	unsigned long port, _value;
	uint8_t value;
	int ret;

	gen_parse_2ul(args, &port, &_value);
	if (rel)
		port += iobase;
	value = (uint8_t) _value;
	ret = d_outb(port, value);
	if (ret) {
		fprintf(stderr, "d_outb error: %s\n", strerror(-ret));
		return ret;
	}
	return 0;
}

int cmd_inw(const char *args, int rel)
{
	unsigned long port;
	uint16_t value;
	int ret;

	gen_parse_1ul(args, &port);
	if (rel)
		port += iobase;
	ret = d_inw(port, &value);
	if (ret) {
		fprintf(stderr, "d_inw error: %s\n", strerror(-ret));
		return ret;
	}
	print_int16(value);
	return 0;
}

int cmd_outw(const char *args, int rel)
{
	unsigned long port, _value;
	uint16_t value;
	int ret;

	gen_parse_2ul(args, &port, &_value);
	if (rel)
		port += iobase;
	value = (uint8_t) _value;
	ret = d_outb(port, value);
	if (ret) {
		fprintf(stderr, "d_outw error: %s\n", strerror(-ret));
		return ret;
	}
	return 0;
}

int cmd_inl(const char *args, int rel)
{
	unsigned long port;
	uint32_t value;
	int ret;

	gen_parse_1ul(args, &port);
	if (rel)
		port += iobase;
	ret = d_inl(port, &value);
	if (ret) {
		fprintf(stderr, "d_inl error: %s\n", strerror(-ret));
		return ret;
	}
	print_int32(value);
	return 0;
}

int cmd_outl(const char *args, int rel)
{
	unsigned long port, _value;
	uint32_t value;
	int ret;

	gen_parse_2ul(args, &port, &_value);
	if (rel)
		port += iobase;
	value = (uint8_t) _value;
	ret = d_outl(port, value);
	if (ret) {
		fprintf(stderr, "d_outl error: %s\n", strerror(-ret));
		return ret;
	}
	return 0;
}

int cmd_readb(const char *args, int rel)
{
	unsigned long addr;
	uint8_t value;
	int ret;

	gen_parse_1ul(args, &addr);
	if (rel)
		addr += membase;
	ret = d_read(addr, &value, 1);
	if (ret) {
		fprintf(stderr, "d_read error: %s\n", strerror(-ret));
		return ret;
	}
	print_int8(value);
	return 0;
}

int cmd_writeb(const char *args, int rel)
{
	unsigned long addr, _value;
	uint8_t value;
	int ret;

	gen_parse_2ul(args, &addr, &_value);
	if (rel)
		addr += membase;
	value = (uint8_t) _value;
	ret = d_write(addr, &value, 1);
	if (ret) {
		fprintf(stderr, "d_write error: %s\n", strerror(-ret));
		return ret;
	}
	return 0;
}

int cmd_readw(const char *args, int rel)
{
	unsigned long addr;
	uint16_t value;
	int ret;

	gen_parse_1ul(args, &addr);
	if (rel)
		addr += membase;
	ret = d_read(addr, &value, 2);
	if (ret) {
		fprintf(stderr, "d_read error: %s\n", strerror(-ret));
		return ret;
	}
	print_int16(value);
	return 0;
}

int cmd_writew(const char *args, int rel)
{
	unsigned long addr, _value;
	uint16_t value;
	int ret;

	gen_parse_2ul(args, &addr, &_value);
	if (rel)
		addr += membase;
	value = (uint16_t) value;
	ret = d_write(addr, &value, 2);
	if (ret) {
		fprintf(stderr, "d_write error: %s\n", strerror(-ret));
		return ret;
	}
	return 0;
}

int cmd_readl(const char *args, int rel)
{
	unsigned long addr;
	uint32_t value;
	int ret;

	gen_parse_1ul(args, &addr);
	if (rel)
		addr += membase;
	ret = d_read(addr, &value, 4);
	if (ret) {
		fprintf(stderr, "d_read error: %s\n", strerror(-ret));
		return ret;
	}
	print_int32(value);
	return 0;
}

int cmd_writel(const char *args, int rel)
{
	unsigned long addr, _value;
	uint32_t value;
	int ret;

	gen_parse_2ul(args, &addr, &_value);
	if (rel)
		addr += membase;
	value = (uint32_t) _value;
	ret = d_write(addr, &value, 4);
	if (ret) {
		fprintf(stderr, "d_write error: %s\n", strerror(-ret));
		return ret;
	}
	return 0;
}

int cmd_dump(const char *args, int align, int rel)
{
	unsigned long addr, len;
	uint8_t *buff;
	int ret;

	gen_parse_2ul(args, &addr, &len);
	if (rel)
		addr += membase;
	buff = (uint8_t *) malloc(len);
	if (buff == NULL) {
		fprintf(stderr, "debug: Memory allocation failure.\n");
		return -ENOMEM;
	}		
	ret = d_read(addr, buff, len);
	if (ret) {
		fprintf(stderr, "d_read error: %s\n", strerror(-ret));
		return ret;
	}
	if (align)
		dump_area_a(addr, buff, len);
	else
		dump_area(addr, buff, len);
	return 0;
}

int cmd_fill(const char *args, int rel)
{
	unsigned long addr, len, _value;
	uint8_t *buff;
	uint8_t value;
	int ret;

	gen_parse_3ul(args, &addr, &len, &_value);
	if (rel)
		addr += membase;
	value = (uint8_t) _value;
	buff = (uint8_t *) malloc(len);
	if (buff == NULL) {
		fprintf(stderr, "debug: Memory allocation failure.\n");
		return -ENOMEM;
	}
	memset(buff, value, len);	
	ret = d_write(addr, buff, len);
	if (ret) {
		fprintf(stderr, "d_write error: %s\n", strerror(-ret));
		return ret;
	}
	return 0;
}

int cmd_parambase(const char *args)
{
	unsigned long tmp;
	char cmd[64];

	sscanf(args, "%64s %li", cmd, &tmp);
	if (tmp > 36) {
		fprintf(stderr, "Unsupported parambase: %ld.\n", tmp);
		return 1;
	}
	parambase = tmp;
	return 0;
}

int cmd_membase(const char *args)
{
	unsigned long addr;

	gen_parse_1ul(args, &addr);
	membase = addr;
	return 0;
}

int cmd_iobase(const char *args)
{
	unsigned long addr;

	gen_parse_1ul(args, &addr);
	iobase = addr;
	return 0;
}

static int parse_bool(const char *val)
{
	if (!strcmp(val, "yes") || !strcmp(val, "y") || !strcmp(val, "1"))
		return 1;
	return 0;
}

#define PARSE_BOOL_OPT(name, variable)					\
	do {								\
		if (!strcmp(opt, (name))) {				\
			(variable) = parse_bool(val);			\
			return 0;					\
		}							\
	} while (0)

int cmd_set(const char *args)
{
	char cmd[64], opt[64], val[64];

	sscanf(args, "%64s %64s %64s", cmd, opt, val);
	opt[63] = '\0';
	val[63] = '\0';
	PARSE_BOOL_OPT("arg_compat", ac_compat);
	PARSE_BOOL_OPT("dump_align", dump_align);
	PARSE_BOOL_OPT("rel_mem", rel_mem);
	PARSE_BOOL_OPT("rel_io", rel_io);
	PARSE_BOOL_OPT("hide_nonascii", hide_nonascii);
	PARSE_BOOL_OPT("hide_nonprintable", hide_nonprintable);
	return -EINVAL;
}

#undef PARSE_BOOL_OPT

void license(void)
{
	fprintf(stderr, "This program is free software; you can redistribute "
			"it and/or modify\n"
			"it under the terms of the GNU General Public License "
			"as published by\n"
			"the Free Software Foundation; either version 2 of the "
			"License, or\n"
			"(at your option) any later version.\n"
			"\n"
			"This program is distributed in the hope that it will "
			"be useful,\n"
			"but WITHOUT ANY WARRANTY; without even the implied "
			"warranty of\n"
			"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  "
			"See the\n"
			"GNU General Public License for more details.\n"
			"\n"
			"You should have received a copy of the GNU General "
			"Public License\n"
			"along with this program; if not, write to the Free "
			"Software\n"
			"Foundation, Inc., 59 Temple Place, Suite 330, Boston, "
			"MA  02111-1307  USA\n");
}

void help(void)
{
	fprintf(stderr, "Supported commands:\n\n"
			"general: exit, q, quit, help, license, compat, native,"
			" parambase, set\n"
			"i/o: iobase\n"
			"i/o 8-bit: i, ib, inb, o, ob, outb,\n"
			"i/o 16-bit: iw, inw, ow, outw,\n"
			"i/o 32-bit: il, inl, ol, outl,\n"
			"memory: membase\n"
			"memory 8-bit: r, rb, readb, w, wb, writeb,\n"
			"memory 16-bit: rw, readw, ww, writew,\n"
			"memory 32-bit: rl, readl, wl, writel,\n"
			"memory block: d, dump, da, dumpa, f, fill\n");
}

static void show_usage(char *name)
{
	fprintf(stderr,	"usage: %s [options]\n"
			"Options:\n"
			"  -c, --compat		Linux MiniDEBUG compatibility "
			"mode\n"
			"  -h, --help		show this help\n"
			"  -V, --version		show version\n",
			(name == NULL) ? "debug" : name);
	exit(0);
}

static void enable_compat(void)
{
	ac_compat = 1;
	dump_align = 1;
	rel_mem = 1;
	rel_io = 1;
	hide_nonascii = 0;
	hide_nonprintable = 1;
}

static void enable_native(void)
{
	ac_compat = 0;
	dump_align = 0;
	rel_mem = 0;
	rel_io = 0;
	hide_nonascii = 1;
	hide_nonprintable = 1;
}

static void parse_command_line(int argc, char **args)
{
	int c;
	int option_index = 0;

	static struct option long_options[] = {
		{"compat", 0, 0, 'c'},
		{"help", 0, 0, 'h'},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};
	
	while (1) {
		c = getopt_long(argc, args, "chV", long_options,
				&option_index);
		if (c == -1)
			break;
		switch (c) {
			case 'c':
				enable_compat();
				break;
			case 'h':
				show_usage(args[0]);
				break;
			case 'V':
				exit(0);
				break;
			default:
				fprintf(stderr, "Unknown option.\n");
				exit(1);
		}
	}

	if (optind < argc)
		fprintf(stderr, "warning: unknown parameters detected.\n");	
}
		
int main(int argc, char **args)
{
	char *tmp = NULL;
	char cmd[64];

	fprintf(stderr, "debug "PACKAGE_VERSION" Copyright (C) 2005  "
			"Krzysztof Mazur <krzysiek@podlesie.net>\n"
			"debug comes with ABSOLUTELY NO WARRANTY;\n"
			"This is free software, and you are welcome to "
			"redistribute it under \n"
			"certain conditions; type `license' for details.\n");
	parse_command_line(argc, args);
	d_ioinit();
	d_meminit();
	rl_outstream = stderr;
	while (1) {
		if (tmp)
			free(tmp);
		tmp = rl_gets();
		if (tmp == NULL) {
			fprintf(stderr, "\n");
			break;
		}
		if (*tmp == '\0')
			continue;
		sscanf(tmp, "%64s", cmd);
		cmd[63] = '\0';
		if (!strcmp(cmd, "exit") || !strcmp(cmd, "quit") ||
				!strcmp(cmd, "q"))
			break;
		if (!strcmp(cmd, "i") || !strcmp(cmd, "ib") ||
				!strcmp(cmd, "inb")) {
			cmd_inb(tmp, rel_io);
			continue;
		}
		if (!strcmp(cmd, "o") || !strcmp(cmd, "ob") ||
				!strcmp(cmd, "outb")) {
			cmd_outb(tmp, rel_io);
			continue;
		}
		if (!strcmp(cmd, "iw") || !strcmp(cmd, "inw")) {
			cmd_inw(tmp, rel_io);
			continue;
		}
		if (!strcmp(cmd, "ow") || !strcmp(cmd, "outw")) {
			cmd_outw(tmp, rel_io);
			continue;
		}
		if (!strcmp(cmd, "il") || !strcmp(cmd, "inl")) {
			cmd_inl(tmp, rel_io);
			continue;
		}
		if (!strcmp(cmd, "ol") || !strcmp(cmd, "outl")) {
			cmd_outl(tmp, rel_io);
			continue;
		}
		if (!strcmp(cmd, "r") || !strcmp(cmd, "rb") ||
				!strcmp(cmd, "readb")) {
			cmd_readb(tmp, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "w") || !strcmp(cmd, "wb") ||
				!strcmp(cmd, "writeb")) {
			cmd_writeb(tmp, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "ww") || !strcmp(cmd, "writew")) {
			cmd_writew(tmp, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "rw") || !strcmp(cmd, "readw")) {
			cmd_readw(tmp, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "wl") || !strcmp(cmd, "writel")) {
			cmd_writel(tmp, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "rl") || !strcmp(cmd, "readl")) {
			cmd_readl(tmp, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "d") || !strcmp(cmd, "dump")) {
			cmd_dump(tmp, dump_align, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "da") || !strcmp(cmd, "dumpa")) {
			cmd_dump(tmp, 1, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "dn") || !strcmp(cmd, "dumpn")) {
			cmd_dump(tmp, 0, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "f") || !strcmp(cmd, "fill")) {
			cmd_fill(tmp, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "set")) {
			cmd_set(tmp);
			continue;
		}
		if (!strcmp(cmd, "membase")) {
			cmd_membase(tmp);
			continue;
		}
		if (!strcmp(cmd, "iobase")) {
			cmd_iobase(tmp);
			continue;
		}
		if (!strcmp(cmd, "license")) {
			license();
			continue;
		}
		if (!strcmp(cmd, "help")) {
			help();
			continue;
		}
		if (!strcmp(cmd, "compat")) {
			enable_compat();
			continue;
		}
		if (!strcmp(cmd, "native")) {
			enable_native();
			continue;
		}
		if (!strcmp(cmd, "parambase")) {
			cmd_parambase(tmp);
			continue;
		}
		fprintf(stderr, "Unknown command: %s\n", cmd);
	}
	fprintf(stderr, "Bye\n");
	return 0;
}
